from typing import List

from nhlscrapi.games.events import EventType
from nhlscrapi.games.playbyplay import Play

from filters import old_filter, play_event_type_filter

class ShotsContainer:
    def __init__(self, game):
        self.game = game

    def shots(self):
        if not (hasattr(self, '_shots') and self._shots):
            self._shots = {
                'shots': self.shots_by_type_with_participants(EventType.Shot),
                'misses': self.shots_by_type_with_participants(EventType.Miss),
                'blocks': self.shots_by_type_with_participants(EventType.Block),
            }

        return self._shots

    def shots_by_type_with_participants(self, shot_type):
        all_plays = self.game.scrapi_game.plays
        shots = old_filter(play_event_type_filter(shot_type), all_plays)

        for shot in shots:
            event_type: str = str(shot.event)
            if event_type == 'Block':
                shot.shot_taker, shot.blocker = self._blocked_shot_participants(shot)
            elif event_type == 'Miss':
                shot.shot_taker, = self._missed_shot_participants(shot)
            else:
                shot.shot_taker, shot.goalie = self._shot_participants(shot)

            if not shot.json:
                shot.for_team, shot.against_team = self._shot_teams(shot)

        return shots

    def _shot_participants(self, shot: Play):
        participants: List[List[dict]] = self.game.event_participants(shot, ('shooter', 'goalie'))
        shooter_json: dict = participants['shooter'][0]
        goalie_json: dict = participants['goalie'][0] if len(participants['goalie']) else None

        shot_taker: dict = self.game.onice_json_from_scrapi(None, (shooter_json['fullName'],))

        if goalie_json:
            goalie: dict = self.game.onice_json_from_scrapi(None, (goalie_json['fullName'],))
        else:
            goalie = None

        return shot_taker, goalie

    def _blocked_shot_participants(self, shot: Play):
        participants: List[List[dict]] = self.game.event_participants(shot, ('shooter', 'blocker'))

        shooter_json: dict = participants['shooter'][0]
        blocker_json: dict = participants['blocker'][0]

        shot_taker: dict = self.game.onice_json_from_scrapi(None, (shooter_json['fullName'],))
        blocker: dict = self.game.onice_json_from_scrapi(None, (blocker_json['fullName'],))

        return shot_taker, blocker

    def _missed_shot_participants(self, shot: Play):
        participants: List[List[dict]] = self.game.event_participants(shot, ('shooter',))

        shooter_json: dict = participants['shooter'][0]
        shot_taker: dict = self.game.onice_json_from_scrapi(None, (shooter_json['fullName'],))

        return shot_taker,

    def _shot_teams(self, shot):
        for on_ice in shot.on_ice['home']:
            if on_ice['id'] == shot.shot_taker['id']:
                return self.game.teams['home'], self.game.teams['away']

        else:
            return self.game.teams['away'], self.game.teams['home']



from nhlscrapi.games.events import EventType
from nhlscrapi.games.playbyplay import Play

import filters

class HitsContainer:
    def __init__(self, game):
        self.game = game

    def hits(self):
        if not (hasattr(self, '_hits') and self._hits):
            all_plays = self.game.scrapi_game.plays
            self._hits = list(filter(filters.play_event_type_filter(EventType.Hit), all_plays))

            for hit in self._hits:
                hit.hitter, hit.hittee = self._hit_participants(hit)

                if not hit.json:
                    hit.for_team, hit.against_team = self._hit_teams(hit)

        return self._hits

    def _hit_participants(self, hit):
        assert isinstance(hit, Play)
        assert hit.event.event_type == EventType.Hit

        participants = self.game.event_participants(hit, ('hitter', 'hittee'))

        hitter_json = participants['hitter'][0] if len(participants['hitter']) else None
        hittee_json = participants['hittee'][0] if len(participants['hittee']) else None

        if hitter_json:
            assert 'fullName' in hitter_json
            hitter = self.game.onice_json_from_scrapi(None, (hitter_json['fullName'],))
        else:
            hitter = None

        if hittee_json:
            assert 'fullName' in hittee_json
            hittee = self.game.onice_json_from_scrapi(None, (hittee_json['fullName'],))
        else:
            hittee = None

        return hitter, hittee

    def _hit_teams(self, hit):
        for on_ice in hit.on_ice['home']:
            if hit.hitter and on_ice['id'] == hit.hitter['id']:
                return self.game.teams['home'], self.game.teams['away']
            elif hit.hittee and on_ice['id'] == hit.hittee['id']:
                return self.game.teams['away'], self.game.teams['home']

        else:
            return self.game.teams['away'], self.game.teams['home']



from nhlscrapi.games.events import EventType
from nhlscrapi.games.playbyplay import Play

import filters

class FaceoffsContainer:
    def __init__(self, game):
        self.game = game

    def faceoffs(self):
        if not (hasattr(self, '_faceoffs') and self._faceoffs):
            all_plays = self.game.scrapi_game.plays
            self._faceoffs = [play for play in all_plays if play.event.event_type == EventType.FaceOff]

            assert isinstance(self._faceoffs, list)
            assert len(self._faceoffs), 'No faceoffs'

            for faceoff in self._faceoffs:
                faceoff.winner, faceoff.loser = self._faceoff_participants(faceoff)

                if not faceoff.json: # no for_team or against_team:
                    faceoff.for_team, faceoff.against_team = self._faceoff_teams(faceoff)


        return self._faceoffs

    def _faceoff_participants(self, faceoff):
        assert isinstance(faceoff, Play)
        assert faceoff.event.event_type == EventType.FaceOff

        participants = self.game.event_participants(faceoff, ('winner', 'loser'))

        winner_json = participants['winner'][0]
        loser_json = participants['loser'][0]

        assert 'fullName' in winner_json
        assert 'fullName' in loser_json

        winner = self.game.onice_json_from_scrapi(None, [winner_json['fullName']])
        loser = self.game.onice_json_from_scrapi(None, [loser_json['fullName']])

        return winner, loser


    def _faceoff_teams(self, faceoff):
        for on_ice in faceoff.on_ice['home']:
            if on_ice['id'] == faceoff.winner['id']:
                return self.game.teams['home'], self.game.teams['away']

        else:
            return self.game.teams['away'], self.game.teams['home']



from nhlscrapi.games.events import EventType
from nhlscrapi.games.playbyplay import Play

import filters

class PenaltiesContainer:
    def __init__(self, game):
        self.game = game

    def penalties(self):
        if not (hasattr(self, '_penalties') and self._penalties):
            all_plays = self.game.scrapi_game.plays
            self._penalties = list(filter(filters.play_event_type_filter(EventType.Penalty), all_plays))

            for penalty in self._penalties:
                penalty.offender, penalty.drawer, penalty.server = self._penalty_participants(penalty)

                if not penalty.json:
                    penalty.for_team, penalty.against_team = self._penalty_teams(penalty)

        return self._penalties

    def _penalty_participants(self, penalty):
        assert isinstance(penalty, Play)
        assert penalty.event.event_type == EventType.Penalty

        participants = self.game.event_participants(penalty, ('penaltyon', 'drewby', 'servedby'))

        offender_json = participants['penaltyon'][0] if len(participants['drewby']) else None
        drawer_json = participants['drewby'][0] if len(participants['drewby']) else None

        if offender_json:
            offender = self.game.onice_json_from_scrapi(None, (offender_json['fullName'],))
        else:
            offender = None

        if drawer_json:
            drawer = self.game.onice_json_from_scrapi(None, (drawer_json['fullName'],))
        else:
            drawer = None

        if participants['servedby']:
            server = self.game.onice_json_from_scrapi(None, (participants['servedby'][0]['fullName'],))
        else:
            server = None

        return offender, drawer, server

    def _penalty_teams(self, penalty):
        for on_ice in penalty.on_ice['home']:
            if penalty.offender and on_ice['id'] == penalty.offender['id']:
                return self.game.teams['home'], self.game.teams['away']
            elif penalty.server and on_ice['id'] == penalty.server['id']:
                return self.game.teams['home'], self.game.teams['away']

        else:
            return self.game.teams['away'], self.game.teams['home']



from typing import List

from nhlscrapi.games.events import EventType
from nhlscrapi.games.playbyplay import Play

from filters import old_filter, play_event_type_filter

class TurnoversContainer:
    def __init__(self, game):
        self.game = game

    def turnovers(self):
        if not (hasattr(self, '_turnovers') and self._turnovers):
            self._turnovers = {
                'giveaways': self.turnovers_by_type_with_participants(EventType.Giveaway),
                'takeaways': self.turnovers_by_type_with_participants(EventType.Takeaway),
            }

        return self._turnovers

    def turnovers_by_type_with_participants(self, turnover_type):
        all_plays = self.game.scrapi_game.plays
        turnovers = old_filter(play_event_type_filter(turnover_type), all_plays)

        for turnover in turnovers:
            if turnover_type == EventType.Giveaway:
                turnover.giver, = self._giveaway_participants(turnover)
            else:
                turnover.taker, = self._takeaway_participants(turnover)

            if not turnover.json:
                turnover.for_team, turnover.against_team = self._turnover_teams(turnover)

        return turnovers

    def _giveaway_participants(self, turnover: Play):
        participants: List[List[dict]] = self.game.event_participants(turnover, ('playerid',))
        player_json: dict = participants['playerid'][0]

        player: dict = self.game.onice_json_from_scrapi(None, (player_json['fullName'],))

        return player,

    def _takeaway_participants(self, turnover: Play):
        participants: List[List[dict]] = self.game.event_participants(turnover, ('playerid',))
        player_json: dict = participants['playerid'][0]

        player: dict = self.game.onice_json_from_scrapi(None, (player_json['fullName'],))

        return player,

    def _turnover_teams(self, turnover):
        for on_ice in turnover.on_ice['home']:
            if hasattr(turnover, 'giver') and on_ice['id'] == turnover.giver['id']:
                return self.game.teams['home'], self.game.teams['away']

            elif hasattr(turnover, 'taker') and on_ice['id'] == turnover.taker['id']:
                return self.game.teams['home'], self.game.teams['away']

        else:
            return self.game.teams['away'], self.game.teams['home']



import re
import collections
from itertools import starmap, chain

from nhlscrapi.games.game import Game, GameKey, GameType
from nhlscrapi.games.events import EventType

from nhlscrapi.games.playbyplay import Play

import filters

from nhlapi import NhlApi

from gametime import game_time, period_time

from containers.hits import HitsContainer
from containers.faceoffs import FaceoffsContainer
from containers.goals import GoalsContainer
from containers.shots import ShotsContainer
from containers.turnovers import TurnoversContainer
from containers.penalties import PenaltiesContainer
from containers.shootouts import ShootoutContainer
from containers.shift import ShiftSummary

GAME_TYPES = [None, GameType.PreSeason, GameType.Regular, GameType.Playoffs]
GAME_TYPE_DIVISOR = 10000


# We do a kind of player lookup against the game roster. There's no way to do this
# before 2010. The best way I can think to manage player lookups is to have the user
# of this library provide a function for the lookups, and that would only be used
# in games where the JSON API cannot be used.
class GameContainer():
    def __init__(self, season, game, json={}):
        print(season, game)
        self.season = int(season)
        self.game = int(game)

        self.nhl_api = NhlApi(self.season, self.game, json=json)
        assert isinstance(self.nhl_api, NhlApi)

        self.scrapi_game = self.create_scrapi_game()
        assert isinstance(self.scrapi_game, Game)

        self.teams = self.nhl_api.live_json()['gameData']['teams']
        self.times = self.nhl_api.live_json()['gameData']['datetime']

        self._assign_meta()
        self._assign_on_ice()

        # This is regrettable
        self.faceoffs_container = FaceoffsContainer(self)
        self.faceoffs = self.faceoffs_container.faceoffs # link these methods
        self.faceoffs()

        self.hits_container = HitsContainer(self)
        self.hits = self.hits_container.hits
        self.hits()

        self.goals_container = GoalsContainer(self)
        self.goals = self.goals_container.goals
        self.goals() # Initialize

        self.shots_container = ShotsContainer(self)
        self.shots = self.shots_container.shots
        self.shots() # Initialize

        self.turnovers_container = TurnoversContainer(self)
        self.turnovers = self.turnovers_container.turnovers
        self.turnovers()

        self.penalties_container = PenaltiesContainer(self)
        self.penalties = self.penalties_container.penalties
        self.penalties()

        self.shootout_container = ShootoutContainer(self)
        self.shootout = self.shootout_container.shootout
        self.shootout()

        self.shifts()

        play_types = set()
        for play in self.scrapi_game.plays:
            play_types.add(str(play.event))

    def create_scrapi_game(self):
        game_type = scrapi_game_type(self.game)
        game_number = scrapi_game_number(self.game)
        scrapi_game_key = GameKey(self.season, game_type, game_number)

        return Game(scrapi_game_key, cum_stats={})

    def players(self):
        return self.nhl_api.players_json().values()

    def rosters(self):
        def augment_player(player):
            player.json = self.onice_json_from_scrapi(None, (player['name'],))
            return player

        if not (hasattr(self, '_rosters') and self._rosters):
            self._rosters = {
                'home': list(map(augment_player, self.scrapi_game.rosters.home_skaters)),
                'away': list(map(augment_player, self.scrapi_game.rosters.away_skaters)),
            }

        return self._rosters

    # Doesn't work. 201620047 has issues with Di Giuseppe showing up as GIUSEPPE DI
    def shifts(self):
        if not (hasattr(self, '_shifts') and self._shifts):
            home_shifts = self.scrapi_game.toi.home_shift_summ
            away_shifts = self.scrapi_game.toi.away_shift_summ

            self._shifts = {
                'home': [ShiftSummary(shift_summary, self) for _, shift_summary in home_shifts.items()],
                'away': [ShiftSummary(shift_summary, self) for _, shift_summary in away_shifts.items()],
            }

            # for player_number, shift_summary in chain(home_shifts.items(), away_shifts.items()):
            #     assert hasattr(shift_summary, 'player_name')
            #     assert 'first' in shift_summary.player_name
            #     assert 'last' in shift_summary.player_name
            #
            #     player_name = shift_summary.player_name['first'] + ' ' + shift_summary.player_name['last']
            #     shift_summary.player = self.onice_json_from_scrapi(player_number, [player_name])
            #
            #     for shift in shift_summary.shifts:
            #         shift['state'] = self._find_state(shift['period'], period_time(shift['end']))
            #
            # self._shifts = list(home_shifts.items()) + list(away_shifts.items())

        return self._shifts

    def _transform_play(self, play):
        assert isinstance(play, Play)
        assert hasattr(play, 'home_on_ice')
        assert isinstance(play.home_on_ice, dict)

        assert hasattr(play, 'vis_on_ice')
        assert isinstance(play.vis_on_ice, dict)

        play.on_ice = {
            'home': list(starmap(self.onice_json_from_scrapi, play.home_on_ice.items())),
            'away': list(starmap(self.onice_json_from_scrapi, list(play.vis_on_ice.items()))),
            #'bucket': calculate_bucket(play)
        }

        for num, info in zip(list(play.vis_on_ice.keys()), list(play.vis_on_ice.values())):
            player = self.onice_json_from_scrapi(num, info)
            play.on_ice['away'].append(player)

        if 'team' in play.json:
            play.for_team = self.teams['home'] if self.teams['home']['id'] == play.json['team']['id'] else self.teams['away']
            play.against_team = self.teams['away'] if self.teams['home']['id'] == play.json['team']['id'] else self.teams['home']

        return play

    def onice_json_from_scrapi(self, number, info):
        assert len(info)
        assert isinstance(info[0], str)

        games_players = list(self.players())

        if len(info) > 2:
            player_name = '-'.join(reversed(info[:-1]))
        else:
            player_name = info[0]

        info_first_name, info_last_name = name_string_split(player_name)

        for game_player in games_players:
            name_match = game_player['fullName'].upper() == player_name.upper()
            silly_match_attempt = game_player['firstName'].upper() in player_name and game_player['lastName'].upper() in player_name

            # reverse of the silly match
            silly_match_reverse = info_first_name in game_player['fullName'].upper() and info_last_name in game_player['fullName'].upper()

            if name_match or silly_match_attempt or silly_match_reverse:
                match = game_player
                break

        else:
            # absolute last case scenario - we really don't want to get here
            # collect all with last name
            last_name_matches = list(filter(lambda p: p['lastName'].upper() == info_last_name, games_players))
            best_match = None
            matches = -1
            for last_name_match in last_name_matches:
                overlap = len([i for i,j in zip(last_name_match['lastName'], info_last_name) if i.upper() == j.upper()])
                if overlap > matches:
                    best_match = last_name_match

            match = best_match

        if len(info) >= 2 and match:
            match['position'] = info[-1]

        return match

    def play_types_match(self, play, play_json):
        json_play_type = re.sub('\s+', '', play_json['result']['event'].lower())
        scrapi_play_type = str(play.event).lower()

        return json_play_type.startswith(scrapi_play_type)

    def _assign_meta(self):
        all_plays = self.scrapi_game.plays

        json_events = list(self.nhl_api.pbp_json()[1:])
        json_plays = filter(is_period_start, json_events)

        for play in all_plays:
            for play_json in json_events:
                if game_time(play_json['about']['period'], play_json['about']['periodTime']) == game_time(play.period, play.time) and self.play_types_match(play, play_json):
                    play.coordinates = play_json['coordinates']
                    play.id = int(play_json['about']['eventId'])

                    play.json = play_json
                    play.type = play_json['result']['event'].lower()
                    break

            else:
                play.json = {}
                play.type = str(play.event).lower()

    def _assign_on_ice(self):
        all_plays = self.scrapi_game.plays
        for play in all_plays:
            play.on_ice = self.scrapi_onice_to_json(play)

            if 'team' in play.json:
                play.for_team = self.teams['home'] if self.teams['home']['id'] == play.json['team']['id'] else \
                self.teams['away']
                play.against_team = self.teams['away'] if self.teams['home']['id'] == play.json['team']['id'] else \
                self.teams['home']

    def scrapi_onice_to_json(self, play):
        home_scrapi_onice = play.home_on_ice
        away_scrapi_onice = play.vis_on_ice
        return {
            'home': [o for o in [self.onice_json_from_scrapi(num, info) for num, info in home_scrapi_onice.items()] if o],
            'away': [o for o in [self.onice_json_from_scrapi(num, info) for num, info in away_scrapi_onice.items()] if o],
        }

    def _find_state(self, period, time):
        for play in reversed(self.scrapi_game.plays):
            if play.period == period and period_time(play.time) < time:
                return play.state

    def event_players_from_type(self, event_players, player_type):
        # In the interest of speed, player_type_match returns a curried function that
        # checks if the types are the same.
        # You know what I had before? I had this:
        # return [player['player'] for player in event_players if player['playerType'].lower() == player_type][0]
        # But I've heard that's not as fast, and I believe it.
        # I don't find this any less readable.
        #   1. Filter the players by type, which returns an iterator
        #   2. Grab the first one
        #   3. Grab the 'player' value from it

        matches = list(filter(filters.player_type_match(player_type), event_players))
        return [match['player'] for match in matches]

    def event_participants(self, event, participant_types):
        assert isinstance(participant_types, collections.Iterable)
        assert isinstance(event, Play)

        event_json = event.json

        participants = {}
        for participant_type in participant_types:
            if event_json:
                participant_matches = self.event_players_from_type(event_json['players'], participant_type)
                participants[participant_type] = participant_matches
            elif hasattr(event.event, 'participants'):
                participant_matches = self.get_participant_old(event, participant_type)
                participants[participant_type] = participant_matches

        return participants


    def get_participant_old(self, play, participant_type: str, basic=False) -> dict:
        participants: list = []

        if participant_type == 'goalie':
            other_participant: str = (self.get_participant_old(play, 'shooter', True) or self.get_participant_old(play, 'scorer', True))[0][1]
            if play.event.home_abbr in other_participant['team']:
                on_ices = play.on_ice['away']
            else:
                on_ices = play.on_ice['home']

            for on_ice in on_ices:
                if on_ice['position'].startswith('G'):
                    participants.append(on_ice)

        else:
            for participant in play.event.participants:
                participant_name: str = participant['name'].upper()

                if participant['playerType'].lower() == participant_type.lower():
                    on_ice_roster = []
                    other_roster = []
                    if participant['num'] in play.home_on_ice: on_ice_roster += play.on_ice['home']
                    if participant['num'] in play.vis_on_ice: on_ice_roster += play.on_ice['away']

                    added_at_least_one = False
                    for on_ice in on_ice_roster:
                        on_ice_name: str = on_ice['lastName'].upper()

                        if on_ice_name == participant_name:
                            if not basic:
                                participants.append(on_ice)
                            else:
                                participants.append((on_ice, participant))

                            added_at_least_one = True

                    if not added_at_least_one:
                        for player in self.players():
                            if player['lastName'].upper() == participant_name:
                                if not basic:
                                    participants.append(player)
                                else:
                                    participants.append((player, participant))
                                break

        return participants


def scrapi_game_type(game):
    game_number_int = int(game) # could be a string
    game_type_digit = game_number_int // GAME_TYPE_DIVISOR

    return GAME_TYPES[game_type_digit]

def scrapi_game_number(game):
    game_number_int = int(game) # could be a string
    game_number = game_number_int % GAME_TYPE_DIVISOR

    return game_number

def is_period_start(play):
    assert 'result' in play
    assert 'event' in play['result']

    play_type = play['result']['event'].lower()

    is_period_play = 'period' in play_type
    is_period_start = 'period start' in play_type
    is_period_end = 'period end' in play_type

    return (not is_period_play) or is_period_start or is_period_end

def name_string_split(name):
    first_name, *rest = name.split()
    last_name = ' '.join(rest)

    return first_name, last_name



from typing import List

from nhlscraper import GameContainer
from utilities import iterator_nonempty

from nhlscrapi.games import events, playbyplay

test_games = (
    # (2016, 20001),
    # (2014, 20010),
    # (2010, 20123), # 2010 is the earliest we can go back with the JSON method for faceoffs
    # # I'm assuming they have JSON rosters going back, though. Should be able to cross reference
    # (2016, 30414),
    # (2016, 20021),
     (2016, 20047),
    (2016, 20046),
    (2016, 20049),
    (2016, 20070),
    (2007, 20008),
    (2016, 21216),
    (2016, 20738),
    (2016, 20428),
)

class TestGame:
    def __init__(self, season, game):
        self.game = GameContainer(season, game)

    def test(self):
        self.test_teams()
        self.test_hits()
        self.test_faceoffs()
        self.test_goals()
        self.test_shots()
        self.test_turnovers()
        self.test_plays()

    def test_plays(self):
        for play in self.game.scrapi_game.plays:
            if 'team' in play.json:
                assert hasattr(play, 'for_team')
                assert 'id' in play.for_team

                assert hasattr(play, 'against_team')
                assert 'id' in play.against_team

                assert play.for_team['id'] != play.against_team['id']

                json_on_ice = len(play.on_ice['home']) + len(play.on_ice['away'])
                scrapi_on_ice = len(play.home_on_ice) + len(play.vis_on_ice)
                assert json_on_ice == scrapi_on_ice

    def test_teams(self):
        assert hasattr(self.game, 'teams')
        assert isinstance(self.game.teams, dict)

        assert 'away' in self.game.teams
        assert isinstance(self.game.teams['away'], dict)
        assert 'id' in self.game.teams['away']
        assert isinstance(self.game.teams['away']['id'], int)
        assert 'franchise' in self.game.teams['away']
        assert isinstance(self.game.teams['away']['franchise'], dict)
        assert 'franchiseId' in self.game.teams['away']['franchise']
        assert isinstance(self.game.teams['away']['franchise']['franchiseId'], int)

        assert 'home' in self.game.teams
        assert isinstance(self.game.teams['home'], dict)
        assert 'id' in self.game.teams['home']
        assert isinstance(self.game.teams['home']['id'], int)
        assert 'franchise' in self.game.teams['home']
        assert isinstance(self.game.teams['home']['franchise'], dict)
        assert 'franchiseId' in self.game.teams['home']['franchise']
        assert isinstance(self.game.teams['home']['franchise']['franchiseId'], int)

    def test_rosters(self):
        assert hasattr(self.game, 'rosters')
        assert callable(self.game.rosters)

        assert 'home' in self.game.rosters()
        assert 'away' in self.game.rosters()

        for player in self.game.rosters()['home']:
            assert hasattr(player, 'json')
            assert isinstance(player['json'], dict)

            self.test_player(player['json'])

        for player in self.game.rosters()['away']:
            assert hasattr(player, 'json')
            assert isinstance(player['json'], dict)

            self.test_player(player['json'])

    def test_faceoffs(self):
        assert hasattr(self.game, 'faceoffs'), 'No faceoffs method'
        assert callable(self.game.faceoffs), 'No faceoffs method'
        assert iterator_nonempty(self.game.faceoffs()), 'No faceoffs'

        faceoffs_json = self.game.nhl_api.faceoffs_json()
        #assert iterator_nonempty(faceoffs_json), 'No faceoffs from JSON API' # Gotta dump these because some game don't have them

        for faceoff in self.game.faceoffs():
            assert isinstance(faceoff, playbyplay.Play)
            assert isinstance(faceoff.event, events.FaceOff)
            assert hasattr(faceoff, 'time')
            assert isinstance(faceoff.time, dict)

            faceoff_time = faceoff.time
            assert 'min' in faceoff_time
            assert 'sec' in faceoff_time

            assert hasattr(faceoff, 'winner')
            self.test_player(faceoff.winner)

            assert hasattr(faceoff, 'loser')
            self.test_player(faceoff.loser)

            assert hasattr(faceoff, 'for_team')
            assert 'id' in faceoff.for_team

            assert hasattr(faceoff, 'against_team')
            assert 'id' in faceoff.for_team

        first_faceoff = self.game.faceoffs()[0]
        assert first_faceoff.time['min'] == 0
        assert first_faceoff.time['sec'] == 0

    def test_hits(self):
        assert hasattr(self.game, 'hits'), 'No hits method'
        assert callable(self.game.hits), 'No hits method'

        hits = self.game.hits()

        assert len(hits), 'no hits'

        for hit in self.game.hits():
            assert isinstance(hit, playbyplay.Play)
            assert isinstance(hit.event, events.Hit)

            assert hasattr(hit, 'hitter')
            assert self.test_player(hit.hitter)

            assert hasattr(hit, 'hittee')
            assert self.test_player(hit.hittee)

            assert hasattr(hit, 'for_team')
            assert 'id' in hit.for_team

            assert hasattr(hit, 'against_team')
            assert 'id' in hit.against_team

    def test_goals(self):
        assert hasattr(self.game, 'goals'), 'No goals method'
        assert callable(self.game.goals), 'No goals method'

        goals = self.game.goals()

        assert len(goals)

        for goal in goals:
            assert isinstance(goal, playbyplay.Play)
            assert isinstance(goal.event, events.Goal)

            assert hasattr(goal, 'shot_taker')
            assert self.test_player(goal.shot_taker)

            assert hasattr(goal, 'assists')
            assert len(goal.assists) in (0,1,2)

            for assist in goal.assists:
                self.test_player(assist)

            assert hasattr(goal, 'goalie')
            if goal.goalie:
                self.test_player(goal.goalie)

            assert hasattr(goal, 'for_team')
            assert 'id' in goal.for_team

            assert hasattr(goal, 'against_team')
            assert 'id' in goal.against_team

    def test_shots(self):
        assert hasattr(self.game, 'shots'), 'No shots method'
        assert callable(self.game.shots), 'No shots method'

        shots: List[playbyplay.Play] = self.game.shots()

        for miss in shots['misses']:
            assert hasattr(miss, 'shot_taker')
            assert isinstance(miss.shot_taker, dict)
            assert 'id' in miss.shot_taker

            assert hasattr(miss, 'for_team')
            assert 'id' in miss.for_team

            assert hasattr(miss, 'against_team')
            assert 'id' in miss.for_team

        for block in shots['blocks']:
            assert hasattr(block, 'shot_taker')
            assert isinstance(block.shot_taker, dict)
            assert self.test_player(block.shot_taker)

            assert hasattr(block, 'blocker')
            assert isinstance(block.blocker, dict)
            assert self.test_player(block.blocker)

            assert hasattr(block, 'for_team')
            assert 'id' in block.for_team

            assert hasattr(block, 'against_team')
            assert 'id' in block.for_team

        for shot in shots['shots']:
            assert hasattr(shot, 'shot_taker')
            assert isinstance(shot.shot_taker, dict)
            assert self.test_player(shot.shot_taker)

            assert hasattr(shot, 'goalie')
            assert isinstance(shot.goalie, dict)
            assert self.test_player(shot.goalie)

            assert hasattr(shot, 'for_team')
            assert 'id' in shot.for_team

            assert hasattr(shot, 'against_team')
            assert 'id' in shot.for_team

    def test_turnovers(self):
        assert hasattr(self.game, 'turnovers'), 'No turnovers method'
        assert callable(self.game.turnovers), 'No turnovers method'

        turnovers: List[playbyplay.Play] = self.game.turnovers()

        for giveaway in turnovers['giveaways']:
            assert hasattr(giveaway, 'giver')
            assert isinstance(giveaway.giver, dict)
            assert self.test_player(giveaway.giver)

            assert hasattr(giveaway, 'for_team')
            assert 'id' in giveaway.for_team

            assert hasattr(giveaway, 'against_team')
            assert 'id' in giveaway.for_team

        for takeaway in turnovers['takeaways']:
            assert hasattr(takeaway, 'taker')
            assert isinstance(takeaway.taker, dict)
            assert self.test_player(takeaway.taker)

            assert hasattr(takeaway, 'for_team')
            assert 'id' in takeaway.for_team

            assert hasattr(takeaway, 'against_team')
            assert 'id' in takeaway.for_team

    def test_penalties(self):
        assert hasattr(self.game, 'penalties'), 'No penalties method'
        assert callable(self.game.penalties), 'No penalties method'

        penalties = self.game.penalties()

        assert len(penalties)

        for penalty in penalties:
            assert isinstance(penalty, playbyplay.Play)
            assert isinstance(penalty.event, events.penalty)

            assert hasattr(penalty, 'offender')
            assert self.test_player(penalty.offender)

            assert hasattr(penalty, 'drawer')
            assert self.test_player(penalty.drawer)

            assert hasattr(penalty, 'for_team')
            assert 'id' in penalty.for_team

            assert hasattr(penalty, 'against_team')
            assert 'id' in penalty.for_team

    def test_shootout(self):
        assert hasattr(self.game, 'shots'), 'No shots method'
        assert callable(self.game.shots), 'No shots method'

        shots: List[playbyplay.Play] = self.game.shootout()

        for shot in shots:
            print(shot)
            assert hasattr(shot, 'shot_taker')
            assert isinstance(shot.shot_taker, dict)
            assert self.test_player(shot.shot_taker)

            assert hasattr(shot, 'goalie')
            assert isinstance(shot.goalie, dict)
            assert self.test_player(shot.goalie)

            assert hasattr(shot, 'for_team')
            assert 'id' in shot.for_team

            assert hasattr(shot, 'against_team')
            assert 'id' in shot.for_team

        print('shoootout')

    def test_player(self, player):
        assert isinstance(player, dict)

        assert 'id' in player
        assert isinstance(player['id'], int)

        assert 'firstName' in player
        assert isinstance(player['firstName'], str)

        assert 'lastName' in player
        assert isinstance(player['lastName'], str)

        assert 'birthDate' in player
        assert isinstance(player['birthDate'], str) # I'd like to change this

        return True


for game in test_games:
    test = TestGame(game[0], game[1])
    test.test()



import re

from nhlscrapi.games.playbyplay import Play

import gametime

def old_filter(f, iter):
    return list(filter(f, iter))

def json_play_match(play):
    assert hasattr(play, 'time')

    def json_match(play_json):
        assert 'about' in play_json
        assert 'periodTime' in play_json['about']
        assert re.match('\d+:\d+', play_json['about']['periodTime'])

        play_time = gametime.period_time(play.time)

        json_min, json_sec = list(map(int, play_json['about']['periodTime'].split(':')))
        json_time = gametime.period_time({ 'min': json_min, 'sec': json_sec })

        return play_time == json_time and play.period == play_json['about']['period']

    return json_match

def play_event_type_filter(event_type):
    def event_type_match(play):
        assert isinstance(play, Play)

        return play.event.event_type == event_type

    return event_type_match

def json_event_type_filter(event_type):
    def event_type_match(event):
        assert isinstance(event, dict)

        return event['result']['event'].upper() == event_type

    return event_type_match

def player_type_match(player_type):
    assert isinstance(player_type, str)

    def match_check(player):
        return player['playerType'].lower() == player_type.lower()

    return match_check



from nhlscrapi.games.events import EventType
from nhlscrapi.games.playbyplay import Play

import filters

class GoalsContainer:
    def __init__(self, game):
        self.game = game

    def goals(self):
        if not (hasattr(self, '_goals') and self._goals):
            all_plays = self.game.scrapi_game.plays
            self._goals = list(filter(filters.play_event_type_filter(EventType.Goal), all_plays))

            for goal in self._goals:
                goal.shot_taker, goal.assists, goal.goalie = self._goal_participants(goal)

                if not goal.json:
                    goal.for_team, goal.against_team = self._goal_teams(goal)

        return self._goals

    def _goal_participants(self, goal):
        assert isinstance(goal, Play)
        assert goal.event.event_type == EventType.Goal

        participants = self.game.event_participants(goal, ('scorer', 'goalie', 'assist'))

        shooter_json = participants['scorer'][0]
        goalie_json = participants['goalie'][0] if len(participants['goalie']) else None
        assists_json = participants['assist']

        if shooter_json:
            shooter = self.game.onice_json_from_scrapi(None, (shooter_json['fullName'],))
        else:
            shooter = None

        if goalie_json:
            goalie = self.game.onice_json_from_scrapi(None, (goalie_json['fullName'],))
        else:
            goalie = None

        assists = [self.game.onice_json_from_scrapi(None, (assist_json['fullName'],)) for assist_json in assists_json]

        return shooter, assists, goalie

    def _goal_teams(self, goal):
        for on_ice in goal.on_ice['home']:
            if on_ice['id'] == goal.shot_taker['id']:
                return self.game.teams['home'], self.game.teams['away']

        else:
            return self.game.teams['away'], self.game.teams['home']



from typing import List

from nhlscrapi.games.events import EventType
from nhlscrapi.games.playbyplay import Play

class ShootoutContainer:
    def __init__(self, game):
        self.game = game

    def shootout(self):
        if not (hasattr(self, '_shootout') and self._shootout):
            shootout_events = filter(is_shootout_event, self.game.scrapi_game.plays)
            for shootout_event in shootout_events:
                shootout_event.shot_taker, shootout_event.goalie = self._shootout_participants(shootout_event)

                if not shootout_event.json:
                    shootout_event.for_team, shootout_event.against_team = self._shootout_teams(shootout_event)

    def _shootout_participants(self, shootout: Play):
        participants: List[List[dict]] = self.game.event_participants(shootout, ('shooter', 'goalie'))
        shooter_json: dict = participants['shooter'][0]
        goalie_json: dict = participants['goalie'][0] if len(participants['goalie']) else None

        shot_taker: dict = self.game.onice_json_from_scrapi(None, (shooter_json['fullName'],))

        if goalie_json:
            goalie: dict = self.game.onice_json_from_scrapi(None, (goalie_json['fullName'],))
        else:
            goalie = None

        return shot_taker, goalie

    def _shootout_teams(self, shootout):
        for on_ice in shootout.on_ice['home']:
            if on_ice['id'] == shootout.goalie['id']:
                return self.game.teams['away'], self.game.teams['home']

        else:
            return self.game.teams['home'], self.game.teams['away']

def is_shootout_event(event):
    return event.event.event_type_str in ('ShootOutAtt', 'ShootOutGoal')



import re

def game_time(period, time):
    return (int(period)-1)*60*20 + period_time(time)

def period_time(time):
    if isinstance(time, dict):
        assert 'min' in time
        assert 'sec' in time

        return time['min'] * 60 + time['sec']

    elif isinstance(time, str):
        minutes, seconds = re.findall('(\d+):(\d+)', str(time))[0]
        return int(minutes)*60 + int(seconds)





from nhlscrapi.games.toi import ShiftSummary as ShiftSummarpi

from gametime import game_time, period_time

class ShiftSummary:
    def __init__(self, shift_summary: ShiftSummarpi, game):
        self.shift_summary = shift_summary
        self.game = game

        player_name = shift_summary.player_name['first'] + ' ' + shift_summary.player_name['last']
        self.player = game.onice_json_from_scrapi(shift_summary.player_num, [player_name])


    @property
    def shifts(self):
        if not (hasattr(self, '_shifts') and self._shifts):
            self._shifts = [Shift(shift, self) for shift in self.shift_summary.shifts]

        return self._shifts


class Shift:
    def __init__(self, shift: dict, parent_summary: ShiftSummary):
        self.shift = shift
        self.parent_summary = parent_summary
        self.game = game = parent_summary.game

        self.player = parent_summary.player
        # self.state = game._find_state(shift['period'], period_time(shift['end'])) Shifts have no concept of "state" as the NHL defines them
        self.shift_number = shift['shift_num']

        self.start = game_time(shift['period'], shift['start'])
        self.end = game_time(shift['period'], shift['end'])
        self.duration = abs(self.end - self.start)

    @property
    def events(self):
        if not(hasattr(self, '_events') and self._events):
            self._events = [event for event in self.game.scrapi_game.plays
                            if self.start < game_time(event.period, event.time) <= self.end
                            or (self.start == game_time(event.period, event.time) and event.event.event_type_str.upper() == 'FACEOFF')
                            ]

        return self._events



from setuptools import setup, find_packages
from pip.req import parse_requirements

requirements = parse_requirements('./requirements.txt', session=False)

setup(
  name="nhlscraper",
  description='NHL Scraper API for Python',
  url='https://gitlab.com/hrel3/nhlscraper',
  author='Justin Barber',
  author_email='justin@hrel.ca',
  python_requires='>=3.6',
  packages=find_packages(),
  version='0.1.1',
  include_package_data=True,
  # Dependent packages (distributions)
  dependency_links=['git+https://github.com/leluso/nhlscrapi.git#egg=nhlscrapier'],
  install_requires=[str(requirement.req) for requirement in requirements],
)



import requests
import filters

#SSSSGGGGGG
LIVE_URL = 'https://statsapi.web.nhl.com/api/v1/game/{season}{game}/feed/live?site=en_nhl'

FACEOFF = 'FACEOFF'
HIT = 'HIT'
GOAL = 'GOAL'
SHOT = 'SHOT'
MISSED_SHOT = 'MISSED SHOT'
BLOCKED_SHOT = 'BLOCKED SHOT'
GIVEAWAY = 'GIVEAWAY'
TAKEAWAY = 'TAKEAWAY'
PENALTY = 'PENALTY'

class NhlApi():
    def __init__(self, season, game, json={}):
        self.season = season
        self.game = game
        self._live_json = json

    def live_json(self):
        if not (hasattr(self, '_live_json') and self._live_json):
            live_url = self.format_url(LIVE_URL)
            live_data_request = requests.get(live_url)

            self._live_json = live_data_request.json()

        return self._live_json

    def pbp_json(self):
        return self.live_json()['liveData']['plays']['allPlays']

    def faceoffs_json(self):
        if not (hasattr(self, '_faceoffs_json') and self._faceoffs_json):
            self._faceoffs_json = list(self.filter_events_by_type(FACEOFF))

        return self._faceoffs_json

    def hits_json(self):
        if not (hasattr(self, '_hits_json') and self._hits_json):
            self._hits_json = self.filter_events_by_type(HIT)

        return self._hits_json

    def goals_json(self):
        if not (hasattr(self, '_goals_json') and self._goals_json):
            self._goals_json = self.filter_events_by_type(GOAL)

        return self._goals_json

    def shots_json(self):
        if not (hasattr(self, '_shots_json') and self._shots_json):
            self._shots_json = self.filter_events_by_type(SHOT)

        return self._shots_json

    def misses_json(self):
        if not (hasattr(self, '_misses_json') and self._misses_json):
            self._misses_json = self.filter_events_by_type(MISSED_SHOT)

        return self._misses_json

    def blocks_json(self):
        if not (hasattr(self, '_blocks_json') and self._blocks_json):
            self._blocks_json = self.filter_events_by_type(BLOCKED_SHOT)

        return self._blocks_json

    def giveaways_json(self):
        if not (hasattr(self, '_giveaways_json') and self._giveaways_json):
            self._giveaways_json = self.filter_events_by_type(GIVEAWAY)

        return self._giveaways_json

    def takeaways_json(self):
        if not (hasattr(self, '_takeaways_json') and self._takeaways_json):
            self._takeaways_json = self.filter_events_by_type(TAKEAWAY)

        return self._takeaways_json

    def penalties_json(self):
        if not (hasattr(self, '_penalties_json') and self._penalties_json):
            self._penalties_json = self.filter_events_by_type(PENALTY)

        return self._penalties_json

    # This isn't actually very good because the team information is wildly useless
    def players_json(self):
        return self.live_json()['gameData']['players']

    def format_url(self, url, old_style=False):
        six_digit_game = str(self.game).zfill(6)

        formatted_season = str(self.season)
        if old_style: # eg: 20152016
            formatted_season += str(self.season+1)

        return url.format(season=formatted_season, game=six_digit_game)

    def filter_events_by_type(self, event_type):
        return list(filter(filters.json_event_type_filter(event_type), self.pbp_json()))
